/*
 * Copyright (c) 2001 University of Bologna
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.genie;

import anthill.*;
import anthill.gf.*;
import anthill.util.*;
import gajit.*;
import gajit.util.*;

/**
 *  Basic implementation of the GeneticFramework interface.
 *
 *  @author Alberto Montresor
 *  @version $REVISION$
 */
public class GeneticFrameworkImpl
extends Printable
implements GeneticFramework
{
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Number of genetic codes constituting the population */
  private int     numcodes;
  
  /** Number of genes (bit sequences) constituting a genetic code */
  private int     numgenes;
  
  /** Size of a single gene */
  private int     sizegenes;
  
  /** Mutation rate */ 
  private double  mutRate;

  /** Percentage of population carried forward unchanged from each generation */
  private double  eliteRate;
  
  /** Percentage deemed unfit for reproduction */
  private double  cullRate;
  
  /** Number of generations to be computed */
  private int generations;
  
  /** Factory of Experiments on which genetic codes are to be evaluted */
  private ExperimentFactory factory;

  /** Population of evolved genetic codes */
  private Population people;
  
  /** Fittest genetic code */
  private GeneticCode bestCode;
  private double bestFitness;
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Instantiates a new genetic framework
   */
  public GeneticFrameworkImpl(int numcodes, int numgenes, int sizegenes, double mutRate,
    double eliteRate, double cullRate, int generations, ExperimentFactory factory)
  {
    this.numcodes = numcodes;
    this.numgenes = numgenes; 
    this.sizegenes = sizegenes; 
    this.mutRate = mutRate; 
    this.eliteRate = eliteRate; 
    this.cullRate = cullRate; 
    this.generations = generations;
    this.factory = factory;
    
    // Initialize genetic codes population
    people = new Population(numcodes, sizegenes, numgenes, numgenes);
    people.setEliteRate(eliteRate);
    people.setCullRate(cullRate);
    people.addOp( new XOverPtOp( 1 ), 1 );
    if (mutRate > 0.0)
      people.addOp( new MutOp( mutRate ), 1 );

    // Initialize statistics
    bestCode = null;
    bestFitness = 0.0;
  }
  
  /**
   *  Tries to evolve a population of genetic codes, using the
   *  configuration parameters contained in this instance.
   *
   *  @param out  print stream used to log information about the
   *    evolution; if print stream is null, no log file is generated.
   */
  public void evolve()
  {
    // Compute generations
    for (int i = 0; i < generations; i++) {
      long time = System.currentTimeMillis();
      
      // Initialize Experiment
      Experiment experiment = factory.getExperiment();

      // Iterate over the population of genetic codes
      Iterator censo = people.iteratorNewOnly();
      while (censo.hasNext()) {
        ChromItem tmp = (ChromItem) censo.next();
        GeneticCode code = new GeneticCode(tmp.getChrom().toBits());

        experiment.addGeneticCode(code);
        experiment.evaluate();
        double val = experiment.getFitness();
        experiment.clear();
        tmp.setFitness(val);
        if (val > bestFitness) {
          bestFitness = val;
          bestCode = code;
        }
      }

      // Compute the next generation
      if (i < generations - 1) {
        people.newGeneration();
      }

      // Print statistics
      Log.println(100, i + " ; " + (System.currentTimeMillis() - time) + " ; " +
          bestFitness + " ; " + bestCode);
    }
  }

  /**
   *  Return the population of evolved genetic codes. 
   */
  public Population getPopulation()
  {
    return people;
  }

  
  /**
   *  Return the genetic code with the best fitness.
   */
  public GeneticCode getFittest()
  {
    return bestCode;
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Accessors / mutators
  ////////////////////////////////////////////////////////////////////////////////////////////
  
 
  /** 
   * Set the mutation rate 
   */ 
  public void setMutRate(double mutRate)
  {
    this.mutRate = mutRate;
    /* FIXME: In order to do this, we need to modify package gajit
    if (mutRate > 0.0)
      people.addOp( new MutOp( mutRate ), 1 );
    */
  }

  /** 
   * Set the percentage of population carried forward unchanged from each generation 
   */
  public void setEliteRate(double eliteRate)
  {
    this.eliteRate = eliteRate;
    people.setEliteRate(eliteRate);
  }
  
  /** 
   * Set the percentage deemed unfit for reproduction 
   */
  public void setCullRate(double cullRate)
  {
    this.cullRate = cullRate;
    people.setCullRate(cullRate);
  }
  
  /** 
   * Set the factory of Experiments on which genetic codes are to be evaluted 
   */
  public void setFactory (ExperimentFactory sf)
  {
    this.factory = factory;
  }
  
  /** 
   * Set the number of genetic codes constituting the population 
   */
  public int getNumcodes()
  {
    return numcodes;
  }
  
  /** 
   * Set the number of genes (bit sequences) constituting a genetic code 
   */
  public int getNumgenes()
  {
    return numgenes;
  }
  
  /** 
   * Set the size of a single gene 
   */
  public int getSizegenes()
  {
    return sizegenes;
  }
  
  /** 
   * Set the mutation rate 
   */ 
  public double getMutRate()
  {
    return mutRate;
  }

  /** 
   * Set the percentage of population carried forward unchanged from each generation 
   */
  public double getEliteRate()
  {
    return eliteRate;
  }
  
  /** 
   * Set the percentage deemed unfit for reproduction 
   */
  public double getCullRate()
  {
    return cullRate;
  }
  
  /** 
   * Set the factory of Experiments on which genetic codes are to be evaluted 
   */
  public ExperimentFactory setFactory()
  {
    return factory;
  }

} // END GeneticFramework

