/*
 * Copyright (c) 2001 University of Bologna
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.nestor;

import anthill.*;
import java.util.*;

/**
 *  LocalStorage is a simple implementation of the Storage interface, whose
 *  aim is to serve as resource repository for anthills during the off-line
 *  simulation. Each storage is characterized by its capacity, i.e. by the
 *  number of documents that can be simultaneously maintained in it.
 *  Due to the fact that the capacity is limited, resources are stored
 *  following a last-recently-used (LRU) policy.
 *
 *  This data structure is designed for single-threaded off-line simulation,
 *  i.e. its methods are not synchronized.
 *
 *  @author Alberto Montresor
 *  @version $REVISION$
 */
public class LocalStorage
implements Storage
{
  
 
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Reference to the storage factory which created this object */
  LocalStorageFactory storagef;
  
  /** Current size (i.e., number of resources) maintained in this storage */
  int size; 
  
  /** LRU queue */
  LinkedList list;
  
  /** Hash map for obtaining resources */
  HashMap map;
  

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Builds a new storage
   */
  public LocalStorage(LocalStorageFactory storagef)
  {
    this.storagef = storagef;
    size = 0;
    list = new LinkedList();
    map = new HashMap();
    
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Storage
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Returns the Resource object associated to the resource identified
   *  by the given name, if such resource is present in this storage;
   *  returns null otherwise. If the resource is present, it is moved
   *  from its current position in the LRU queue to the first one.
   *
   *  @param name The identifier of the resource
   *  @return the resource to be found
   */
  public Resource[] getResources(String query)
  {
    Resource resource = (Resource) map.get(query);
    if (resource != null) {
      list.remove(query);
      list.addFirst(query);
      return new Resource[] { resource };
    } else {
      return null;
    }
  }

  /**
   *  Add a resource to the storage. This method is invoked 
   *  from an anthill on behalf of an ant. If the resource is 
   *  already present in the storage, it is moved from its 
   *  current position in the LRU queue to the first position.
   *  If the resource is not present, it is stored in the
   *  storage and it is placed in the first position of the
   *  LRU queue. In both cases, the storage verifies that
   *  the number of resources in the system does not exceed
   *  its current capacity; if so, resources in the last 
   *  positions of the LRU queue are removed from the storage.
   *  <p>
   *  The current capacity of the storage is retrieved by
   *  interrogating the storage factory which created this
   *  object. In this way, it is possible to experiment
   *  systems in which the capacity of the storage is dynamic.
   *
   *  @param resource the resource to be added
   */
  public void addAntResource(Resource resource)
  {
    // Retrieve the name of the resource
    String name = resource.getName();
    
    // If the resource is present, remove it from its current LRU position
    if (list.remove(name))
      size--;
    
    // Add the resource to the head of the LRU list
    list.addFirst(name);
    map.put(name, resource);
    size++;

    // Remove resources exceeding capacity
    int capacity = storagef.getCapacity();
    while (size >= capacity) {
      map.remove(list.removeLast());
      size--;
    }
  }
  
  /**
   *  Add a resource to the storage. This method is invoked 
   *  from an anthill on behalf of an anthill client. In this 
   *  simple implementation, the semantics of a local insertion 
   *  is equal to thesemantics of a general insertion.
   *
   *  @param resource the resource to be added
   */
  public void addClientResource(Resource resource)
  {
    addAntResource(resource);
  }
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Clearable
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Restore the object to its initial state.
   */
  public void clear()
  {
    size = 0;
    map.clear();
    list.clear();
  }

} // END LocalStorage

