/*
 * Copyright (c) 2001 University of Bologna
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library Public License for more details.
 *
 * You should have received a copy of the GNU Library Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.nestor;

import java.util.*;
import anthill.*;

/**
 *  A Anthill class collects the components of a Anthill site: (i) a AntView
 *  object, which is the view of a Anthill for an Ant; (ii) a Gate object,
 *  which manage the communication between anthills; (iii) an Manager,
 *  i.e. the object which manage the activities performed by different agents.
 *  <p>
 *  Implementation note: this implementation is supposed to be used as
 *  simple implementation of <CODE>Anthill</CODE> in the off-line
 *  environment. Here, we assume that the Manager object used is single-threaded;
 *  in this way, we avoid the costs of synchronized methods.
 *
 *  @author Alberto Montresor
 *  @version $REVISION$
 */
public class NestorAnthill
implements Anthill, AntListener
{
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Anthill factory to get information about factories */
  private NestorAnthillFactory anthillf;
  
  /** Reference to the resource database */
  private Storage storage;
  
  /** Communication handler to be used to communicate with other anthills */
  private Gate gate;
  
  /** Communication handler to be used to communicate with other anthills */
  private QuerySet queries;

  /** Thread manager needed to be used to manage concurrent agents */
  private Manager manager;

  /** Neighbours hash table, with AnthillId as key */
  private HashMap map;

  /** Neighbours array */
  private Neighbour[] neighbours;
  
  /** GeneticCode */
  private GeneticCode code;
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Constructs a new anthill, by putting together a Storage
   *  (a resource database), a Gate (i.e. the basic API for 
   *  communicating with other anthills) and an Manager
   *  (i.e., a thread handler managing the agent execution).
   *  <p>
   *  The constructor has package visibility to avoid
   *  instances not builded through the appropriate factory.
   */
  NestorAnthill(NestorAnthillFactory anthillf)
  {
    this.anthillf = anthillf;
    init(true);
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from AntListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Receive an agent from outside and deliver it to the
   *  to the activity manager.
   */
  public void deliver(AntDescriptor desc)
  {
    AntView view = 
      anthillf.getAntViewFactory().getAntView(this, desc);
    manager.add(view);
  }
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Anthill
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Return the identifier of this Anthill
   */
  public AnthillId getId()
  {
    return gate.getId();
  }
  
  /**
   *  Return the Manager associated to this Anthill
   */
  public Manager getManager()
  {
    return manager;
  }

  /**
   *  Accept a resource request originated by an anthill client.
   *  The resource to be searched is identified by name. The method
   *  returns immediately. When the resource is found, a Resource
   *  object will be returned to the specified resource listener.
   *  <p>
   *  Implementation note: If the resource can be found locally, 
   *  the listener is immediately contacted. Otherwise, a new 
   *  <CODE>Ant</CODE> is generated and is delivered to the ant 
   *  manager.
   *
   *  @param query The identifier of the resource to be searched
   *  @param listener The listener object which will be notified
   *    when the resource has been found.
   */
  public void request(String query, QueryListener listener)
  {
    Resource[] resources = storage.getResources(query);
    if (resources != null) { // The resource is stored locally
      
      listener.returnResources(resources);
      
    } else { // The resource is not here
    
      // Prepare the Ant
      int ttl = anthillf.getTTL();
      Ant ant = anthillf.getAntFactory().getAnt(code, ttl);
      ant.setQuery(query);
      AntDescriptor desc = new AntDescriptor(null, ant, ttl);
      
      // Deliver the ant to the ant manager
      AntView view = 
        anthillf.getAntViewFactory().getAntView(this, desc);
      manager.add(view);
      queries.add(query, listener);
    }
  }
  
  /**
   *  Insert a new resource in this anthill. Delagates
   *  the insertion to the storage.
   * 
   *  @param resource the resource to be inserted.
   */
  public void insert(Resource resource)
  {
    storage.addAntResource(resource);
  }

  
  /**
   *  Add the identifier of a new neighbour anthill to the set of 
   *  neighbours known to this <CODE>Anthill</CODE>, on behalf of 
   *  an Anthill client. 
   *
   *  @param aid neighbour identifier
   */
  public void addClientNeighbour(AnthillId aid)
  {
    map.put(aid, new Neighbour(aid));
    neighbours = null;
  }
  
  
  /**
   *  Add the identifier of a new neighbour anthill to the set of 
   *  neighbours known to this <CODE>Anthill</CODE>, on behalf of 
   *  an ant. 
   *
   *  @param aid neighbour identifier
   */
  public void addAntNeighbour(AnthillId aid)
  {
    map.put(aid, new Neighbour(aid));
    neighbours = null;
  }

  
  /**
   * Remove a neighbour identifier from the set of neighbour
   * @param aid neighbour identifier
   */
  public void removeNeighbour(AnthillId aid)
  {
    map.remove(aid);
    neighbours = null;
  }
  
  /**
   * Return the neighbours known to this Anthill
   */
  public Neighbour[] getNeighbours()
  {
    if (neighbours == null) {
      neighbours = new Neighbour[map.size()];
      Iterator iterator = map.values().iterator();
      for (int i=0; i < neighbours.length; i++) 
        neighbours[i] = (Neighbour) iterator.next();
    }
    return neighbours;
  }
  
  /**
   *  Add a new genetic code to the set of genetic codes known to this anthill.
   *  When creating a new ant, an anthill may choose from one of the
   *  available genetic codes. This implementation of <CODE>Anthill</CODE> 
   *  does not enable the use of more than one genetic code; thus, duplicated
   *  invocations of <CODE>addGeneticCode()</CODE> throws an 
   *  <CODE>IllegalStateException</CODE>.
   * 
   *  @param code the genetic code to be added
   *  @throws IllegalStateException if no other genetic code can be added to
   *    the anthill
   */
  public void addGeneticCode(GeneticCode code)
  {
    if (this.code == null)
      this.code = code;
    else
      throw new IllegalStateException("Maximum one genetic code in " + 
        this.getClass().getName());
  }
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Package methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Return the <CODE>Gate</CODE> object used by this <CODE>Anthill</CODE>
   */
  Gate getGate()
  {
    return gate;
  }
    
  /**
   *  Return the <CODE>Storage</CODE> object used by this <CODE>Anthill</CODE>
   */
  Storage getStorage()
  {
    return storage;
  }
    
  /**
   *  Return the <CODE>QuerySet</CODE> object used by this <CODE>Anthill</CODE>
   */
  QuerySet getQuerySet()
  {
    return queries;
  }
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Clearable
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Restore the object to its initial state.
   */
  public void clear()
  {
    init(false);
  }

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Initialize the object
   *
   *  @param constructor true if the method is invoked by the
   *    constructor
   */
  private void init(boolean constructor)
  {
    /*
     *  Init the anthill data structures
     */
    if (constructor) {
      // Creates the data structure needed by the Anthill
      storage = anthillf.getStorageFactory().getStorage();
      gate = anthillf.getGateFactory().getGate();
      queries = anthillf.getQuerySetFactory().getQuerySet();
      manager = anthillf.getManagerFactory().getManager();
    } else {
      storage.clear();
      gate.clear();
      manager.clear();
      queries.clear();
    }
    
    /*
     * Init the neighbour list (initial size = 10)
     */
    neighbours = new Neighbour[0];
    if (constructor) {
      map = new HashMap();
    } else {
      map.clear();
    }
    
    code = null;

    /* 
     * Notify the communication handler that this object will receive 
     * agents from outside 
     */
    gate.setAntListener(this);
    
    
  }    
  
  
} // END NestorAnthill
